
/*******************************************************************************
* Copyright (c) 2016 Olivier Langella <Olivier.Langella@moulon.inra.fr>.
*
* This file is part of peptider.
*
*     peptider is free software: you can redistribute it and/or modify
*     it under the terms of the GNU General Public License as published by
*     the Free Software Foundation, either version 3 of the License, or
*     (at your option) any later version.
*
*     peptider is distributed in the hope that it will be useful,
*     but WITHOUT ANY WARRANTY; without even the implied warranty of
*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*     GNU General Public License for more details.
*
*     You should have received a copy of the GNU General Public License
*     along with peptider.  If not, see <http://www.gnu.org/licenses/>.
*
* Contributors:
*     Olivier Langella <Olivier.Langella@moulon.inra.fr> - initial API and implementation
******************************************************************************/

#include "fastasource.h"
#include <QDebug>
#include <QThreadPool>
#include <pappsomspp/fasta/fastareader.h>
#include "digestionpipeline.h"
#include "digestproductmatcher.h"
#include "secondpass.h"
#include "../utils/peptiderparams.h"

FastaSource::FastaSource(BasePsmReporter & reporter): _reporter(reporter)
{
    const PeptiderParams & params = PeptiderParams::Instance();

    _minimum_first_pass_evalue = params.get(PeptiderParamPappsoDouble::MaximumPeptideEvalueThreshold);
    _second_pass = params.get(PeptiderParamBool::SecondPassUse);
    _second_pass_peptide_evalue_selection = params.get(PeptiderParamPappsoDouble::SecondPassPeptideEvalueSelection);
}

FastaSource::~FastaSource()
{
}


void FastaSource::add(const QFileInfo & fasta_file) {
    _fasta_file_list.push_back( {(std::int8_t)_fasta_file_list.size(), fasta_file,false});
}

void FastaSource::addDecoy(const QFileInfo & fasta_decoy_file) {
    _fasta_file_list.push_back( {(std::int8_t)_fasta_file_list.size(), fasta_decoy_file,true});
}


void FastaSource::setSequence(const QString& description, const QString& sequence) {
    //qDebug() << "FastaSource::setSequence " << description << " " <<sequence;
    //ProteinSp protein_sp = Protein(description, new_seq.replace(QRegExp("\\*$"), "")).makeProteinSp();

    //_kinase.eat(protein_sp,*_p_peptide_size2varmod);

    //QThreadPool::globalInstance()->start(new ProteinToSpectrumProcess(_p_mzdata, _p_digestion_pipeline, description, sequence));


    // QMutexLocker lock(&_mutex);


    if (_p_matcher_thread->size() > _max_protein_list_size) {
        while (QThreadPool::globalInstance()->activeThreadCount() >= QThreadPool::globalInstance()->maxThreadCount()) {
            QThread::currentThread()->sleep (1);
        }
        QThreadPool::globalInstance()->start(_p_matcher_thread);
        _p_matcher_thread = new DigestProductMatcher(_p_mzdata, true);
    }

    _p_matcher_thread->push_back_protein_to_digest(_current_sequence_database, description, sequence);
    //digestProtein(protein_sp);

}

void FastaSource::scan(MzData * p_mzdata) {

    _p_mzdata = p_mzdata;
    _reporter.setFastaDatabaseList(_fasta_file_list);
    _p_matcher_thread = new DigestProductMatcher(_p_mzdata, true);


    //read fasta proteins, launch 1 thread on each protein to digest, modify and score it against the spectrum collection
    p_mzdata->buildSpectrumCollectorList();

    for (FastaDatabase & fasta_database: _fasta_file_list) {
        _current_sequence_database = fasta_database;
        QFile fasta_file (fasta_database.sequence_database_file_info.absoluteFilePath());
        fasta_file.open(QIODevice::ReadOnly);
        pappso::FastaReader reader(*this);
        reader.parse(&fasta_file);
        fasta_file.close();
    }

    qDebug() << "FastaSource::scan _p_digestion_thread->size() " ;
    QThreadPool::globalInstance()->start(_p_matcher_thread);
    QThreadPool::globalInstance()->waitForDone();

    _p_mzdata->reduceFirstPass(_minimum_first_pass_evalue);

    if (_second_pass) {
        // SecondPass second_pass_identifier( _p_mzdata);

        std::vector<SpectrumDataCollectorSp>::const_iterator it = p_mzdata->beginSpectrumDataCollector();

        std::vector<SpectrumDataCollectorSp>::const_iterator itend = p_mzdata->endSpectrumDataCollector();

// 	PeptideStore peptide_store;
        _p_matcher_thread = new DigestProductMatcher(_p_mzdata, false);


        //sort identified proteins :


        struct SecondPassProtein {
            std::int8_t sequence_database_id;
            pappso::ProteinSp protein_sp;
            bool reverse;
        };
        std::vector<SecondPassProtein> second_pass_protein_list;
        while (it != itend) {
            //spectrum->print(_reporter);
            std::vector<PsmScore> result_list = it->get()->getPsmScoreListBeneathEvalue(_second_pass_peptide_evalue_selection);
            qDebug() << "FastaSource::scan 2d pass result_list.size() " << result_list.size() << " scan num " << it->get()->getQualifiedSpectrum().getSpectrumId().getScanNum();
            for (PsmScore & score: result_list) {
                if (!score.digest_product.reverse) {
                    second_pass_protein_list.push_back( { score.digest_product.sequence_database_id, score.digest_product.protein_sp, score.digest_product.reverse});
                }
            }
            it++;
        }
        qDebug() << "FastaSource::scan 2d pass second_pass_protein_list.size() before " << second_pass_protein_list.size() ;

        std::sort( second_pass_protein_list.begin(), second_pass_protein_list.end(), [](const SecondPassProtein & a, const SecondPassProtein &  b) {
            if (a.protein_sp.get() == b.protein_sp.get()) {
                return a.reverse < b.reverse;
            }
            return a.protein_sp.get() < b.protein_sp.get();
        } );

        second_pass_protein_list.erase(std::unique(second_pass_protein_list.begin(), second_pass_protein_list.end(), [](const SecondPassProtein & a, const SecondPassProtein &  b) {
            if ((a.reverse == b.reverse) && (a.protein_sp.get() == b.protein_sp.get()))  {
                return true;
            }
            return false;
        }), second_pass_protein_list.end());

        qDebug() << "FastaSource::scan 2d pass second_pass_protein_list.size() " << second_pass_protein_list.size() ;

        for (SecondPassProtein & _second_pass_protein: second_pass_protein_list) {
            _p_matcher_thread->pushBackProteinSecondPass(_second_pass_protein.sequence_database_id, _second_pass_protein.protein_sp, _second_pass_protein.reverse);

            if (_p_matcher_thread->secondPassSize() > 0) {
                QThreadPool::globalInstance()->start(_p_matcher_thread);

                //qDebug() << "FastaSource::scan 2d pass  waitForDone";
                //QThreadPool::globalInstance()->waitForDone();
                _p_matcher_thread = new DigestProductMatcher(_p_mzdata, false);
            }

        }
        QThreadPool::globalInstance()->start(_p_matcher_thread);
        qDebug() << "FastaSource::scan 2d pass  waitForDone";
        QThreadPool::globalInstance()->waitForDone();
        _p_mzdata->reduceFirstPass(_minimum_first_pass_evalue);
    }

    _reporter.writeResults(_p_mzdata);





}

