
/*******************************************************************************
* Copyright (c) 2016 Olivier Langella <Olivier.Langella@moulon.inra.fr>.
*
* This file is part of peptider.
*
*     peptider is free software: you can redistribute it and/or modify
*     it under the terms of the GNU General Public License as published by
*     the Free Software Foundation, either version 3 of the License, or
*     (at your option) any later version.
*
*     peptider is distributed in the hope that it will be useful,
*     but WITHOUT ANY WARRANTY; without even the implied warranty of
*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*     GNU General Public License for more details.
*
*     You should have received a copy of the GNU General Public License
*     along with peptider.  If not, see <http://www.gnu.org/licenses/>.
*
* Contributors:
*     Olivier Langella <Olivier.Langella@moulon.inra.fr> - initial API and implementation
******************************************************************************/

#include "peptider.h"
#include <QThread>
#include <QThreadPool>
#include <QDebug>
#include <QCommandLineParser>
#include <QDateTime>
#include <QTimer>
#include <QFileInfo>
#include <pappsomspp/pappsoexception.h>
#include "core/mzdata.h"
#include "core/fastasource.h"
#include "reporter/odspsmreporter.h"
#include "reporter/mzidentmlreporter.h"
#include <odsstream/odsdocwriter.h>
#include "utils/peptiderparams.h"

Peptider::Peptider(QObject *parent) :
    QObject(parent)
{
    // get the instance of the main application
    app = QCoreApplication::instance();
    // setup everything here
    // create any global objects
    // setup debug and warning mode
}

// 10ms after the application starts this method will run
// all QT messaging is running at this point so threads, signals and slots
// will all work as expected.
void Peptider::run()
{
    // ./src/pt-fastagrouper ../../pappsomspp/test/data/asr1_digested_peptides.txt



    //./src/pt-fastarenamer -i /gorgone/pappso/jouy/database/Strepto_Aaron/20151106_grouping_ortho_uniprot_TK24_M145.fasta -c /gorgone/pappso/jouy/database/Strepto_Aaron/20151106_grouping_TK24_M145.ods -o /tmp/test.fasta

    QTextStream errorStream(stderr, QIODevice::WriteOnly);
    QTextStream outputStream(stdout, QIODevice::WriteOnly);

    try {
        qDebug() << "Peptider::run() begin";
        QCommandLineParser parser;

        //throw pappso::PappsoException("test");
        parser.setApplicationDescription(QString(SOFTWARE_NAME).append(" ").append(PEPTIDER_VERSION));
        parser.addHelpOption();
        parser.addVersionOption();
        parser.addPositionalArgument("FASTA files", QCoreApplication::translate("main", "List of Fasta files to use for database search."));
        QCommandLineOption fastaOption(QStringList() << "f" << "fasta",
                                       QCoreApplication::translate("main", "FASTA file <fasta>."),
                                       QCoreApplication::translate("main", "fasta"));
        QCommandLineOption decoyOption(QStringList() << "d" << "decoy",
                                       QCoreApplication::translate("main", "FASTA decoy file <decoy>."),
                                       QCoreApplication::translate("main", "decoy"));
        QCommandLineOption debugScanOption(QStringList()  << "debug",
                                           QCoreApplication::translate("main", "write informations about scans numbers defined in a list (separated by spaces)"),
                                           QCoreApplication::translate("main", "debug"));

        QCommandLineOption outputFileOption(QStringList() << "o" << "output",
                                            QCoreApplication::translate("main", "output file <file> (*.ods or *.mzid)."),
                                            QCoreApplication::translate("main", "file"));


        QCommandLineOption paramFileOption(QStringList() << "p" << "parameters",
                                           QCoreApplication::translate("main", "parameters file <parameters> (*.ods)."),
                                           QCoreApplication::translate("main", "parameters"));


        QCommandLineOption mzOption(QStringList() << "m" << "mz",
                                    QCoreApplication::translate("main", "mz file <mz>."),
                                    QCoreApplication::translate("main", "mz"));

        QCommandLineOption cpusOption(QStringList() << "c" << "cpus",
                                      QCoreApplication::translate("main", "number of CPUs to use <cpus>."),
                                      QCoreApplication::translate("main", "cpus"));
        parser.addOption(cpusOption);
        parser.addOption(mzOption);
        parser.addOption(fastaOption);
        parser.addOption(decoyOption);
        parser.addOption(debugScanOption);
        parser.addOption(outputFileOption);
        parser.addOption(paramFileOption);

        qDebug() << "Peptider::run() 1";

        // Process the actual command line arguments given by the user
        parser.process(*app);

        // QCoreApplication * app(this);
        // Add your main code here
        qDebug() << "Peptider.Run is executing";
        QString paramFileStr = parser.value(paramFileOption);
        PeptiderParams & peptider_params = PeptiderParams::Instance();
        if (paramFileStr.isEmpty()) {
        }
        else {
            QFileInfo param_info(paramFileStr);
            if (param_info.exists()) {
                peptider_params.load(paramFileStr);
            }
            else {
                OdsDocWriter param_doc_writer(paramFileStr);
                peptider_params.save(param_doc_writer);
                param_doc_writer.close();
            }
        }


        const QDateTime dt_begin = QDateTime::currentDateTime();
        const QStringList args = parser.positionalArguments();

        uint cpu_number = 50;
        QString cpusStr = parser.value(cpusOption);
        if (!cpusStr.isEmpty()) {
            cpu_number = cpusStr.toUInt();
        }

        uint ideal_number_of_thread = (uint) QThread::idealThreadCount();
        //QThreadPool::globalInstance()->setMaxThreadCount(1);
        if (cpu_number > ideal_number_of_thread) {
            cpu_number = ideal_number_of_thread;
        }
        else {
            QThreadPool::globalInstance()->setMaxThreadCount(cpu_number);
        }
        outputStream << QObject::tr("using up to %1 threads").arg(QThreadPool::globalInstance()->maxThreadCount()) << endl;
        outputStream.flush();

        QString outputFileStr = parser.value(outputFileOption);
        BasePsmReporter * p_reporter =nullptr;
        OdsDocWriter * p_ods_writer =nullptr;
        if (outputFileStr.isEmpty()) {
            p_ods_writer = new OdsDocWriter("peptider_report.ods");
            p_reporter = new OdsPsmReporter(*p_ods_writer);
        }
        else {
            QFileInfo fi(outputFileStr);
            QString ext = fi.suffix();  // ext = "gz"
            if (ext == "ods") {
                p_ods_writer = new OdsDocWriter(outputFileStr);
                OdsPsmReporter * ods_reporter = new OdsPsmReporter(*p_ods_writer);
                p_reporter = ods_reporter;
                QStringList debugScanStrList =  parser.value(debugScanOption).split(" ", QString::SkipEmptyParts);
                for (QString & scan_str : debugScanStrList) {
                    ods_reporter->addDebugScanIndex(scan_str.toUInt());
                }
            }
            else if (ext == "mzid") {
                p_reporter = new MzIdentMlReporter(outputFileStr);
            }
            else {
                throw pappso::PappsoException(QObject::tr("output file format *.%1 not supported. Available formats are *.ods or *.mzid").arg(ext));
            }
        }

        QString mzFileStr = parser.value(mzOption);

        if (!mzFileStr.isEmpty()) {
            //build a spectrum collection from mz file

            outputStream << QObject::tr("loading spectra") << endl;
            outputStream.flush();
            MzData mz_data(mzFileStr);
            FastaSource fasta_source(*p_reporter);
            QStringList fasta_list = parser.positionalArguments();
            for (QString & fasta_file:fasta_list) {
                fasta_source.add(QFileInfo(fasta_file));
            }

            QString fastaStr = parser.value(fastaOption);
            if (!fastaStr.isEmpty()) {
                fasta_source.add(QFileInfo(fastaStr));
            }
            QString decoyStr = parser.value(decoyOption);
            if (!decoyStr.isEmpty()) {
                fasta_source.addDecoy(QFileInfo(decoyStr));
            }

            outputStream << QObject::tr("peptide identification is running") << endl;
            outputStream.flush();
            fasta_source.scan(&mz_data);
        }

        p_reporter->close();


        if (p_reporter != nullptr) {
            delete p_reporter;
        }
        if (p_ods_writer != nullptr) {
            delete p_ods_writer;
        }

    }
    catch (pappso::PappsoException& error)
    {
        errorStream << "Oops! an error occurred in peptider. Dont Panic :" << endl;
        errorStream << error.qwhat() << endl;
        exit(1);
        app->exit(1);
    }

    catch (std::exception& error)
    {
        errorStream << "Oops! an error occurred in peptider. Dont Panic :" << endl;
        errorStream << error.what() << endl;
        exit(1);
        app->exit(1);
    }

    outputStream << QObject::tr("peptide identification is finished") << endl;
    outputStream.flush();

    // you must call quit when complete or the program will stay in the
    // messaging loop
    quit();
}

// call this routine to quit the application
void Peptider::quit()
{
    // you can do some cleanup here
    // then do emit finished to signal CoreApplication to quit
    emit finished();
}

// shortly after quit is called the CoreApplication will signal this routine
// this is a good place to delete any objects that were created in the
// constructor and/or to stop any threads
void Peptider::aboutToQuitApp()
{
    // stop threads
    // sleep(1);   // wait for threads to stop.
    // delete any objects
}


int main(int argc, char **argv)
{
    //QTextStream consoleErr(stderr);
    //QTextStream consoleOut(stdout, QIODevice::WriteOnly);
    //ConsoleOut::setCout(new QTextStream(stdout, QIODevice::WriteOnly));
    //ConsoleOut::setCerr(new QTextStream(stderr, QIODevice::WriteOnly));
    qDebug() << "main begin";
    QCoreApplication app(argc, argv);
    qDebug() << "main 1";
    QCoreApplication::setApplicationName("peptider");
    QCoreApplication::setApplicationVersion(PEPTIDER_VERSION);
    QLocale::setDefault(QLocale::system());

    // create the main class
    Peptider myMain;
    // connect up the signals
    QObject::connect(&myMain, SIGNAL(finished()),
                     &app, SLOT(quit()));
    QObject::connect(&app, SIGNAL(aboutToQuit()),
                     &myMain, SLOT(aboutToQuitApp()));
    qDebug() << "main 2";


    // This code will start the messaging engine in QT and in
    // 10ms it will start the execution in the MainClass.run routine;
    QTimer::singleShot(10, &myMain, SLOT(run()));
    return app.exec();
}
